{-----------------------------------------------------------------------------
 Unit Name: DynRegClasses
 Author:    Maximiliano Podostroiec (delphi@podostro.com.ar)
 Purpose:   Enables application to use dynamic registration of classes.
            Enables application to interact with list of dynamically registered classes
            Takes off responsability from programmer to add RegisterClass on BPLs
            Allows for applications to use real plug and play BPLs as runtime modules extensor.

 History:
       v1.0 Written and published

 Future:
      Lots of speed-ups
      Cleaner code
      Documents and examples
-----------------------------------------------------------------------------}
unit DynRegClasses;

interface

uses
  SysUtils, Classes;

type

  EDynRegClassNotFound = class(EClassNotFound); // Exception class used by FindDynRegClass
  EDynRegClassInvalidModule = class(EInvalidContainer); // Exception class used by DynRegisterClasses
  EDynRegClassAlreadyExists = class(EFilerError); // Exception class used by DynRegisterClasses

  // Callback function for notifying dynamically registered classes.
  TDynRegClassListNotify = function(const AClass: TClass; var AParam): Boolean;
  TDynRegClassCheckNotify = function(const AClass: TClass): Boolean;

  // Dynamically registered classes replacements for Classes units GetClass/FindClass
  function GetDynRegClass(const ClassName: String): TClass;
  function FindDynRegClass(const ClassName: String): TClass;

  // Dynamically registered classes checks for already registered class
  function IsDynRegClass(const ClassName: String): Boolean; overload;
  function IsDynRegClass(const AClass: TClass): Boolean; overload;

  // Notifier to list dynamic registered classes
  procedure GetDynRegClasses(const DynRegClassListNotify: TDynRegClassListNotify; var AParam);

  // Method for loaded module exploration and dynamic registration of classes
  procedure DynRegisterClasses(const AModule: HMODULE; const DynRegClassCheckNotify: TDynRegClassCheckNotify = nil; const FailIfRegisteredInClassesUnit: Boolean = True);

  // Methods for dispose of dynamically registered classes
  procedure DynUnRegisterClass(const AClass: TClass);
  procedure DynUnRegisterClasses(const AModule: HMODULE);

implementation

uses
  Windows;

type
  TClassModuleRec = record  // Record to save module in which AClass was loaded
    AModule: HMODULE;
    AClass: TClass
  end;
  PClassModuleRec = ^TClassModuleRec;

  TClassesList = class // List to hold TClassModuleRec elements
  private
    FList: TList; // Pointer list points to TClassModuleRec elements
    function Highest: Integer; // Highest element (replaces FList.Count - 1
    procedure DisposeClassModuleRec(const P: Pointer); // Code reusability, dispose of TClassModuleRec element
  protected
    function Get(const Index: Integer): TClass; // Gets TClassModuleRec element
  public
    constructor Create; // Constructor constructs class and FList
    destructor Destroy; override; // Frees FList and disposes all TClassModuleRec elements
    function Add(const AModule: HMODULE; const AClass: TClass): Integer; // Adds a new TClassModuleRec elemente
    function IndexOf(const AClass: TClass): Integer; // Gets index of Class in FList
    function Count: Integer; // Gets count of TClassModuleRec elements
    procedure Delete(const Index: Integer); // Deletes a Class
    procedure DeleteModule(const AModule: HMODULE); // Deletes all classes belonging to AModule
    property Items[const Index: Integer]: TClass read Get; default; // default property accesing classes
  end;

const
  SClassNotFound = 'Dynamic registered class "%s" not found';
  SClassAlreadyExists = 'A class named %s already exists';
  SCODESectionTag = 'CODE';

var
  DynRegClassesList: TClassesList;

{-----------------------------------------------------------------------------
  Procedure: TClassesList.Highest
  Author:    Maximiliano Podostroiec
  Date:      2006-08-18
  Description: Gets highest element
-----------------------------------------------------------------------------}
function TClassesList.Highest: Integer;
begin

  Result:= (FList.Count-1);

end;

{-----------------------------------------------------------------------------
  Procedure: TClassesList.DisposeClassModuleRec
  Author:    Maximiliano Podostroiec
  Date:      2006-08-18
  Description: Disposes of a TClassModuleRec element receiving its unchanging pointer
-----------------------------------------------------------------------------}
procedure TClassesList.DisposeClassModuleRec(const P: Pointer);
var
  X: PClassModuleRec;
begin

  X:= P;
  Dispose(X);

end;

{-----------------------------------------------------------------------------
  Procedure: TClassesList.Get
  Author:    Maximiliano Podostroiec
  Date:      2006-08-18
  Description: Gets Class element from TClassModuleRec element at Index
-----------------------------------------------------------------------------}
function TClassesList.Get(const Index: Integer): TClass;
begin

  Result:= PClassModuleRec(FList[Index])^.AClass;

end;

{-----------------------------------------------------------------------------
  Procedure: TClassesList.Create
  Author:    Maximiliano Podostroiec
  Date:      2006-08-18
  Description: Constructor which constructs object and child FList
-----------------------------------------------------------------------------}
constructor TClassesList.Create;
begin

  inherited;
  FList:= TList.Create;

end;

{-----------------------------------------------------------------------------
  Procedure: TClassesList.Destroy
  Author:    Maximiliano Podostroiec
  Date:      2006-08-18
  Description: Destructor, destroys object and free all TClassModuleRec elements
-----------------------------------------------------------------------------}
destructor TClassesList.Destroy;
var
  I: Integer;
begin

  if Assigned(FList) then begin
    for I:= Highest downto 0 do
      DisposeClassModuleRec(FList[I]);
    FreeAndNil(Flist);
  end;
  inherited;

end;

{-----------------------------------------------------------------------------
  Procedure: TClassesList.Add
  Author:    Maximiliano Podostroiec
  Date:      2006-08-18
  Description: adds a new TClassModuleRec element to list 
-----------------------------------------------------------------------------}
function TClassesList.Add(const AModule: HMODULE; const AClass: TClass): Integer;
var
  P: PClassModuleRec;
begin

  New(P);
  P^.AModule:= AModule;
  P^.AClass:= AClass;
  Result:= FList.Add(P);

end;

{-----------------------------------------------------------------------------
  Procedure: TClassesList.IndexOf
  Author:    Maximiliano Podostroiec
  Date:      2006-08-18
  Description: Gets index of a class by class reference  
-----------------------------------------------------------------------------}
function TClassesList.IndexOf(const AClass: TClass): Integer;
begin

  Result:= Highest;
  while (Result > -1) and (PClassModuleRec(FList[Result])^.AClass <> AClass) do
    Dec(Result);

end;

{-----------------------------------------------------------------------------
  Procedure: TClassesList.Count
  Author:    Maximiliano Podostroiec
  Date:      2006-08-18
  Description: Gets elements count
-----------------------------------------------------------------------------}
function TClassesList.Count: Integer;
begin

  Result:= FList.Count;

end;

{-----------------------------------------------------------------------------
  Procedure: TClassesList.Delete
  Author:    Maximiliano Podostroiec
  Date:      2006-08-18
  Description: Deletes TClassModuleRec element
-----------------------------------------------------------------------------}
procedure TClassesList.Delete(const Index: Integer);
begin

  DisposeClassModuleRec(FList[Index]);
  Flist.Delete(Index);

end;

{-----------------------------------------------------------------------------
  Procedure: TClassesList.DeleteModule
  Author:    Maximiliano Podostroiec
  Date:      2006-08-18
  Description: Deletes all TClassModuleRec elements belonging to AModule 
-----------------------------------------------------------------------------}
procedure TClassesList.DeleteModule(const AModule: HMODULE);
var
  I: Integer;
begin

  I:= Highest;
  while (I > -1) do begin

    if (PClassModuleRec(FList[I])^.AModule = AModule) then begin
      Dispose(FList[I]);
      FList.Delete(I);
    end;
    Dec(I);

  end;


end;

{-----------------------------------------------------------------------------
  Procedure: GetDynRegClass
  Author:    Maximiliano Podostroiec
  Date:      2006-08-17
  Description: Looks up for a dynamically registered class by name (replaces: Classes.GetClass)
-----------------------------------------------------------------------------}
function GetDynRegClass(const ClassName: String): TClass;
var
  I: Integer;
begin

  Result:= nil;
  I:= 0;
  while (Result = nil) and (I < DynRegClassesList.Count) do begin // Stop if found or reached last class
    if (DynRegClassesList[I].ClassName = ClassName) then
      Result:= DynRegClassesList[I]; // Found the class
    Inc(I);
  end;

end;

{-----------------------------------------------------------------------------
  Procedure: FindDynRegClass
  Author:    Maximiliano Podostroiec
  Date:      2006-08-17
  Description: Looks up for a dynamically registered class by name (replaces: Classes.FindClass)
-----------------------------------------------------------------------------}
function FindDynRegClass(const ClassName: String): TClass;
begin

  Result:= GetDynRegClass(ClassName); // Search for the class
  if (Result = nil) then // Didnt found the class
    raise EDynRegClassNotFound.CreateFmt(SClassNotFound, [ClassName]);

end;

{-----------------------------------------------------------------------------
  Procedure: IsDynRegClass
  Author:    Maximiliano Podostroiec
  Date:      2006-08-17
  Description: Checks if exists dynamically registered class - by name 
-----------------------------------------------------------------------------}
function IsDynRegClass(const ClassName: String): Boolean;
begin

  Result:= (GetDynRegClass(ClassName) <> nil); // Search class by name

end;

{-----------------------------------------------------------------------------
  Procedure: IsDynRegClass
  Author:    Maximiliano Podostroiec
  Date:      2006-08-17
  Description: Checks if exists dynamically registered class - by class reference 
-----------------------------------------------------------------------------}
function IsDynRegClass(const AClass: TClass): Boolean;
begin

  Result:= (DynRegClassesList.IndexOf(AClass) >= 0); // Search class by class reference

end;

{-----------------------------------------------------------------------------
  Procedure: GetDynRegClasses
  Author:    Maximiliano Podostroiec
  Date:      2006-08-17
  Description: Notifies list of dynamically registered classes trough callback function 
-----------------------------------------------------------------------------}
procedure GetDynRegClasses(const DynRegClassListNotify: TDynRegClassListNotify; var AParam);
var
  I: Integer;
begin

  I:= 0;
  while (I < DynRegClassesList.Count) and DynRegClassListNotify(DynRegClassesList[I], AParam) do
    Inc(I); // Increment I until no more items on the list, or user is not interested receiving more notifications

end;

{-----------------------------------------------------------------------------
  Procedure: DynRegisterClass
  Author:    Maximiliano Podostroiec
  Date:      2006-08-17
  Description: Adds class to dynamically registered classes list
-----------------------------------------------------------------------------}
procedure DynRegisterClass(AModule: HModule; AClass: TClass; FailIfRegisteredInClassesUnit: Boolean);
begin

  // Class already registered either dynamically or statically(if FailIfRegisteredInClassesUnit is true)
  if IsDynRegClass(AClass) or IsDynRegClass(AClass.ClassName) or (FailIfRegisteredInClassesUnit and (GetClass(AClass.ClassName) <> nil)) then
    raise EDynRegClassAlreadyExists.CreateFmt(SClassAlreadyExists, [AClass.ClassName]) // raise exception
  else begin
    DynRegClassesList.Add(Amodule, AClass); // Add to list
  end;

end;

{-----------------------------------------------------------------------------
  Procedure: DynRegisterClasses (
             Based on Ludovic Duboiss "Classes hack" found on borland.public.delphi.opentoolsapi newsgroup)
             Using less variables
             Avoiding unnecesary use of Exit procedure
             Avoiding use of PPointer type
             Recoded assembly code as pascal for easier understanding
             Increased Reusability of code

  Author:    Maximiliano Podostroiec
  Date:      2006-08-17
  Description: Explores a loaded module and dynamically registers contained classes
-----------------------------------------------------------------------------}
procedure DynRegisterClasses(const AModule: HMODULE; const DynRegClassCheckNotify: TDynRegClassCheckNotify = nil; const FailIfRegisteredInClassesUnit: Boolean = True);
var
  ImageNTHeadersOffset: PImageNtHeaders;
  pStart,
  pEnd,
  p: Pointer;

  // Gets pointer to asked Section start
  function GetSectionHeader(const ASectionName: string): Boolean;
  var
    I: Integer;
  begin

    Result:= False;
    I:= -1; // Initialized with -1 to start while loop with 0
    // Initialized with -1 element offset to start while loop with first element
    pStart:= PImageSectionHeader(Integer(ImageNTHeadersOffset) + Sizeof(TImageNTHeaders) - SizeOf(TImageSectionHeader));
    // Loop until CODE Section found or no more sections to check 
    while (not(Result) and (I < (ImageNTHeadersOffset^.FileHeader.NumberOfSections - 1))) do begin
      Inc(I); // Starts with 0
      Inc(PImageSectionHeader(pStart)); // Starts at first ImageSectionHeader, then skips to next
      // True if current SectionHeaders name = SCODESectionTag
      Result:= (StrLIComp(PChar(@PImageSectionHeader(pStart)^.Name), PChar(ASectionName), IMAGE_SIZEOF_SHORT_NAME) = 0);
    end;

  end;

  // Checks for APointer to be nil or between pMin and pMax
  function InRangeOrNil(APointer, pMin, pMax: Pointer): Boolean;
  begin

    Result:= (APointer = nil) or
             ((Integer(Apointer) >= Integer(pMin)) and
              (Integer(Apointer) <= Integer(pMax)))

  end;

  // Checks if pointer points to a class identifier
  function IsIdent(p: PChar): Boolean;
  var
    I: Integer;
  begin

    Result:= False;
    I:= 1;
    if (Ord(p^) > 0) and ((p+1)^ in ['A'..'Z', 'a'..'z', '_']) then // Is not an empty string and first char is Alphabetic or _
      while (I < Ord(p^)) do begin
        Inc(I); // Starts at 2nd char of string
        Result:= ((p+I)^ in ['0'..'9', 'A'..'Z', 'a'..'z', '_']); // 2 to last char is Alphanumeric or _
      end;

  end;

  // Added below functions to simplify understanding of condictions

  // Checks for valid magic token on module
  function ValidMSEXEModule: Boolean;
  begin

    Result:= (PImageDosHeader(AModule)^.e_magic = IMAGE_DOS_SIGNATURE);

  end;

  // Checks for valid and readable PE Module
  function ValidReadableNTPEModule: Boolean;
  begin

    ImageNTHeadersOffset:= PImageNtHeaders(LongInt(AModule) +  PImageDosHeader(AModule)^._lfanew);
    Result:= (not(IsBadReadPtr(ImageNTHeadersOffset, SizeOf(TImageNTHeaders))) and
             (ImageNTHeadersOffset^.Signature = IMAGE_NT_SIGNATURE));

  end;

  // Checks if pointer is pointing to a VMT
  function isVirtualMethodTable: Boolean;
  begin

    Result:= Integer(p^) = Integer(p) - vmtSelfPtr; 

  end;

  // Checks if VMT elements exists and are in available range
  function InRangeOrNilTableElements: Boolean;

    // Do the real check
    function Sub_InRangeOrNilTableElements(Elements: Array of Integer; pMin: Pointer): Boolean;
    var
      I: Integer;
    begin

      Result:= True;
      I:= 0;
      // Stops if false is reached
      while (Result and (I < High(Elements))) do begin
        Result:= InRangeOrNil(Pointer(Pointer(Integer(p^) + Elements[I])^), pMin, pEnd);
        Inc(I);
      end;

    end;

  const
    ElementsFurtherInCode: Array[0..3] of Integer = (vmtClassName, vmtDynamicTable, vmtMethodTable, vmtFieldTable);
    ElementsAnyPartOfCode: Array[0..3] of Integer = (vmtTypeInfo, vmtInitTable, vmtAutoTable, vmtIntfTable);

  begin

    Result:= Sub_InRangeOrNilTableElements(ElementsFurtherInCode, Pointer(p^)) and
             Sub_InRangeOrNilTableElements(ElementsAnyPartOfCode, pStart);

  end;

  // Checks if vmtClassName points to a valid Class name
  function isClassIdent: Boolean;
  begin

    Result:= IsIdent(PChar(Pointer(Integer(p^) + vmtClassName)^));

  end;

begin

//  No initialization needed, no read done before write.
//
//  ImageNTHeadersOffset:= nil;
//  pStart:= nil;
//  pEnd:= nil;
//  p:= nil;

  // One big condition to know if I must work to avoid unneeded variables usage
  if ValidMSEXEModule and ValidReadableNTPEModule and GetSectionHeader(SCODESectionTag) then begin

    // Setting pEnd before resetting pStart to avoid retaining Section Header pointer in another variable
    pEnd:= Pointer(AModule + PImageSectionHeader(pStart)^.VirtualAddress + (PImageSectionHeader(pStart)^.SizeOfRawData - SizeOf(Pointer)));
    pStart:= Pointer(AModule + PImageSectionHeader(pStart)^.VirtualAddress); // Real CODE Section data start

    p:= pStart;
    while Integer(p) < Integer(pEnd) do begin

      if isVirtualMethodTable and InRangeOrNilTableElements and isClassIdent then begin

        if Assigned(DynRegClassCheckNotify) then
          if not DynRegClassCheckNotify(TClass(p^)) then
            Continue;
        DynRegisterClass(AModule, TClass(p^), FailIfRegisteredInClassesUnit);

      end;
      
      Inc(Integer(p), SizeOf(Pointer)); // Since pCode points to pointers, inc it Sizeof(Pointer) bytes

    end;

  end else
    raise EDynRegClassInvalidModule.Create('Not a valid module. Must be valid PE Module');

end;

{-----------------------------------------------------------------------------
  Procedure: DynUnRegisterClass
  Author:    Maximiliano Podostroiec
  Date:      2006-08-18
  Description: Unregister one dynamically registered class
-----------------------------------------------------------------------------}
procedure DynUnRegisterClass(const AClass: TClass);
begin

  if IsDynRegClass(AClass) then
    DynRegClassesList.Delete(DynRegClassesList.IndexOf(AClass));

end;

{-----------------------------------------------------------------------------
  Procedure: DynUnRegisterClasses
  Author:    Maximiliano Podostroiec
  Date:      2006-08-18
  Description: Unregisters all dynamically registered classes beloging to AModule
-----------------------------------------------------------------------------}
procedure DynUnRegisterClasses(const AModule: HMODULE);
begin

  DynRegClassesList.DeleteModule(AModule);

end;

Initialization
  DynRegClassesList:= TClassesList.Create;

finalization
  FreeAndNil(DynRegClassesList);

end.
